@@ -1,6 +1,6 @@
-#!/usr/bin/env python3.6
+#!/usr/bin/env python3.7
 #
-# Copyright (c) 2013-2021 by Ron Frederick <ronf@timeheart.net> and others.
+# Copyright (c) 2013-2024 by Ron Frederick <ronf@timeheart.net> and others.
 #
 # This program and the accompanying materials are made available under
 # the terms of the Eclipse Public License v2.0 which accompanies this
@@ -33,7 +33,7 @@ class MySSHClientSession(asyncssh.SSHClientSession):
 
 class MySSHClient(asyncssh.SSHClient):
     def connection_made(self, conn: asyncssh.SSHClientConnection) -> None:
-        print('Connection made to %s.' % conn.get_extra_info('peername')[0])
+        print(f'Connection made to {conn.get_extra_info('peername')[0]}.')
 
     def auth_completed(self) -> None:
         print('Authentication successful.')
@@ -46,6 +46,6 @@ async def run_client() -> None:
         await chan.wait_closed()
 
 try:
-    asyncio.get_event_loop().run_until_complete(run_client())
+    asyncio.run(run_client())
 except (OSError, asyncssh.Error) as exc:
     sys.exit('SSH connection failed: ' + str(exc))
