@@ -37,7 +37,7 @@ from asyncssh.socks import SOCKS5, SOCKS5_AUTH_NONE
 from asyncssh.socks import SOCKS4_OK_RESPONSE, SOCKS5_OK_RESPONSE_HDR
 
 from .server import Server, ServerTestCase
-from .util import asynctest, echo, make_certificate
+from .util import asynctest, echo, make_certificate, try_remove
 
 
 def _echo_non_async(stdin, stdout, stderr=None):
@@ -183,6 +183,18 @@ class _TCPAsyncConnectionServer(_TCPConnectionServer):
             return listen_host != 'fail'
 
 
+class _TCPAcceptHandlerServer(Server):
+    """Server for testing forwarding accept handler"""
+
+    async def server_requested(self, listen_host, listen_port):
+        """Handle a request to create a new socket listener"""
+
+        def accept_handler(_orig_host: str, _orig_port: int) -> bool:
+            return True
+
+        return accept_handler
+
+
 class _UNIXConnectionServer(Server):
     """Server for testing direct and forwarded UNIX domain connections"""
 
@@ -258,6 +270,24 @@ class _CheckForwarding(ServerTestCase):
 
         self.assertEqual(b''.join(data), result)
 
+    async def _check_local_connection(self, listen_port, delay=None):
+        """Open a local connection and test if an input line is echoed back"""
+
+        reader, writer = await asyncio.open_connection('127.0.0.1',
+                                                       listen_port)
+
+        await self._check_echo_line(reader, writer, delay=delay)
+
+    async def _check_local_unix_connection(self, listen_path):
+        """Open a local connection and test if an input line is echoed back"""
+
+        # pylint doesn't think open_unix_connection exists
+        # pylint: disable=no-member
+        reader, writer = await asyncio.open_unix_connection(listen_path)
+        # pylint: enable=no-member
+
+        await self._check_echo_line(reader, writer)
+
 
 class _TestTCPForwarding(_CheckForwarding):
     """Unit tests for AsyncSSH TCP connection forwarding"""
@@ -278,14 +308,6 @@ class _TestTCPForwarding(_CheckForwarding):
 
         await self._check_echo_block(reader, writer)
 
-    async def _check_local_connection(self, listen_port, delay=None):
-        """Open a local connection and test if an input line is echoed back"""
-
-        reader, writer = await asyncio.open_connection('127.0.0.1',
-                                                       listen_port)
-
-        await self._check_echo_line(reader, writer, delay=delay)
-
     @asynctest
     async def test_ssh_create_tunnel(self):
         """Test creating a tunneled SSH connection"""
@@ -310,8 +332,8 @@ class _TestTCPForwarding(_CheckForwarding):
     async def test_ssh_connect_tunnel_string(self):
         """Test connecting a tunneled SSH connection via string"""
 
-        async with self.connect(tunnel='%s:%d' % (self._server_addr,
-                                                  self._server_port)) as conn:
+        async with self.connect(tunnel=f'{self._server_addr}:'
+                                f'{self._server_port}') as conn:
             await self._check_connection(conn)
 
     @asynctest
@@ -319,9 +341,8 @@ class _TestTCPForwarding(_CheckForwarding):
         """Test failed connection on a tunneled SSH connection via string"""
 
         with self.assertRaises(asyncssh.ChannelOpenError):
-            await asyncssh.connect('\xff',
-                                   tunnel='%s:%d' % (self._server_addr,
-                                                     self._server_port))
+            await asyncssh.connect(
+                '\xff', tunnel=f'{self._server_addr}:{self._server_port}')
 
     @asynctest
     async def test_proxy_jump(self):
@@ -339,6 +360,23 @@ class _TestTCPForwarding(_CheckForwarding):
         finally:
             os.remove('.ssh/config')
 
+    @asynctest
+    async def test_proxy_jump_multiple(self):
+        """Test connecting a tunnneled SSH connection using ProxyJump"""
+
+        write_file('.ssh/config', 'Host target\n'
+                   '  Hostname localhost\n'
+                   f'  Port {self._server_port}\n'
+                   f'  ProxyJump localhost:{self._server_port},'
+                   f'localhost:{self._server_port}\n'
+                   'IdentityFile ckey\n', 'w')
+
+        try:
+            async with self.connect(host='target', username='ckey'):
+                pass
+        finally:
+            os.remove('.ssh/config')
+
     @asynctest
     async def test_proxy_jump_encrypted_key(self):
         """Test ProxyJump with encrypted client key"""
@@ -400,8 +438,10 @@ class _TestTCPForwarding(_CheckForwarding):
 
         async with self.connect() as conn:
             async with conn.listen_ssh(port=0, server_factory=Server,
-                                       server_host_keys=['skey']) as server2:
-                listen_port = server2.get_port()
+                                       server_host_keys=['skey']) as server:
+                listen_port = server.get_port()
+
+                self.assertEqual(server.get_addresses(), [('', listen_port)])
 
                 async with asyncssh.connect('127.0.0.1', listen_port,
                                             known_hosts=(['skey.pub'], [], [])):
@@ -411,10 +451,9 @@ class _TestTCPForwarding(_CheckForwarding):
     async def test_ssh_listen_tunnel_string(self):
         """Test opening a tunneled SSH listener via string"""
 
-        async with self.listen(tunnel='ckey@%s:%d' % (self._server_addr,
-                                                      self._server_port),
-                               server_factory=Server,
-                               server_host_keys=['skey']) as server:
+        async with self.listen(
+                tunnel=f'ckey@{self._server_addr}:{self._server_port}',
+                server_factory=Server, server_host_keys=['skey']) as server:
             listen_port = server.get_port()
 
             async with asyncssh.connect('127.0.0.1', listen_port,
@@ -426,11 +465,9 @@ class _TestTCPForwarding(_CheckForwarding):
         """Test open failure on a tunneled SSH listener via string"""
 
         with self.assertRaises(asyncssh.ChannelListenError):
-            await asyncssh.listen('\xff',
-                                  tunnel='%s:%d' % (self._server_addr,
-                                                    self._server_port),
-                                  server_factory=Server,
-                                  server_host_keys=['skey'])
+            await asyncssh.listen(
+                '\xff', tunnel=f'{self._server_addr}:{self._server_port}',
+                server_factory=Server, server_host_keys=['skey'])
 
     @asynctest
     async def test_ssh_listen_tunnel_default_port(self):
@@ -584,6 +621,65 @@ class _TestTCPForwarding(_CheckForwarding):
                 await self._check_local_connection(listener.get_port(),
                                                    delay=0.1)
 
+    @asynctest
+    async def test_forward_local_port_accept_handler(self):
+        """Test forwarding of a local port with an accept handler"""
+
+        def accept_handler(_orig_host: str, _orig_port: int) -> bool:
+            return True
+
+        async with self.connect() as conn:
+            async with conn.forward_local_port('', 0, '', 7,
+                                               accept_handler) as listener:
+                await self._check_local_connection(listener.get_port(),
+                                                   delay=0.1)
+
+    @asynctest
+    async def test_forward_local_port_accept_handler_denial(self):
+        """Test forwarding of a local port with an accept handler denial"""
+
+        async def accept_handler(_orig_host: str, _orig_port: int) -> bool:
+            return False
+
+        async with self.connect() as conn:
+            async with conn.forward_local_port('', 0, '', 7,
+                                               accept_handler) as listener:
+                listen_port = listener.get_port()
+
+                reader, writer = await asyncio.open_connection('127.0.0.1',
+                                                               listen_port)
+
+                self.assertEqual((await reader.read()), b'')
+
+                writer.close()
+                await maybe_wait_closed(writer)
+
+    @unittest.skipIf(sys.platform == 'win32',
+                     'skip UNIX domain socket tests on Windows')
+    @asynctest
+    async def test_forward_local_path_to_port(self):
+        """Test forwarding of a local UNIX domain path to a remote TCP port"""
+
+        async with self.connect() as conn:
+            async with conn.forward_local_path_to_port('local', '', 7):
+                await self._check_local_unix_connection('local')
+
+        try_remove('local')
+
+    @unittest.skipIf(sys.platform == 'win32',
+                     'skip UNIX domain socket tests on Windows')
+    @asynctest
+    async def test_forward_local_path_to_port_failure(self):
+        """Test failure forwarding a local UNIX domain path to a TCP port"""
+
+        open('local', 'w').close()
+
+        async with self.connect() as conn:
+            with self.assertRaises(OSError):
+                await conn.forward_local_path_to_port('local', '', 7)
+
+        try_remove('local')
+
     @asynctest
     async def test_forward_local_port_pause(self):
         """Test pause during forwarding of a local port"""
@@ -642,6 +738,18 @@ class _TestTCPForwarding(_CheckForwarding):
                     await conn.forward_local_port('', listener.get_port(),
                                                   '', 7)
 
+    @unittest.skipIf(sys.platform == 'win32',
+                     'skip UNIX domain socket tests on Windows')
+    @asynctest
+    async def test_forward_port_to_path_bind_error(self):
+        """Test error binding a local port forwarding to remote path"""
+
+        async with self.connect() as conn:
+            async with conn.forward_local_port('0.0.0.0', 0, '', 7) as listener:
+                with self.assertRaises(OSError):
+                    await conn.forward_local_port_to_path(
+                            '', listener.get_port(), '')
+
     @asynctest
     async def test_forward_connect_error(self):
         """Test error connecting a local forwarding port"""
@@ -688,6 +796,24 @@ class _TestTCPForwarding(_CheckForwarding):
         server.close()
         await server.wait_closed()
 
+    @unittest.skipIf(sys.platform == 'win32',
+                     'skip UNIX domain socket tests on Windows')
+    @asynctest
+    async def test_forward_remote_port_to_path(self):
+        """Test forwarding of a remote port to a local UNIX domain socket"""
+
+        server = await asyncio.start_unix_server(echo, 'local')
+
+        async with self.connect() as conn:
+            async with conn.forward_remote_port_to_path(
+                    '', 0, 'local') as listener:
+                await self._check_local_connection(listener.get_port())
+
+        server.close()
+        await server.wait_closed()
+
+        try_remove('local')
+
     @asynctest
     async def test_forward_remote_specific_port(self):
         """Test forwarding of a specific remote port"""
@@ -790,6 +916,33 @@ class _TestTCPForwarding(_CheckForwarding):
             await listener.wait_closed()
 
 
+class _TestTCPForwardingAcceptHandler(_CheckForwarding):
+    """Unit tests for TCP forwarding with accept handler"""
+
+    @classmethod
+    async def start_server(cls):
+        """Start an SSH server which supports TCP connection forwarding"""
+
+        return await cls.create_server(
+            _TCPAcceptHandlerServer, authorized_client_keys='authorized_keys')
+
+    @asynctest
+    async def test_forward_remote_port_accept_handler(self):
+        """Test forwarding of a remote port with accept handler"""
+
+        server = await asyncio.start_server(echo, None, 0,
+                                            family=socket.AF_INET)
+        server_port = server.sockets[0].getsockname()[1]
+
+        async with self.connect() as conn:
+            async with conn.forward_remote_port(
+                    '', 0, '127.0.0.1', server_port) as listener:
+                await self._check_local_connection(listener.get_port())
+
+        server.close()
+        await server.wait_closed()
+
+
 class _TestAsyncTCPForwarding(_TestTCPForwarding):
     """Unit tests for AsyncSSH TCP connection forwarding with async return"""
 
@@ -822,16 +975,6 @@ class _TestUNIXForwarding(_CheckForwarding):
 
         await self._check_echo_line(reader, writer, encoded=True)
 
-    async def _check_local_unix_connection(self, listen_path):
-        """Open a local connection and test if an input line is echoed back"""
-
-        # pylint doesn't think open_unix_connection exists
-        # pylint: disable=no-member
-        reader, writer = await asyncio.open_unix_connection(listen_path)
-        # pylint: enable=no-member
-
-        await self._check_echo_line(reader, writer)
-
     @asynctest
     async def test_unix_connection(self):
         """Test opening a remote UNIX connection"""
@@ -891,7 +1034,7 @@ class _TestUNIXForwarding(_CheckForwarding):
             await listener.wait_closed()
             listener.close()
 
-        os.remove('echo')
+        try_remove('echo')
 
     @asynctest
     async def test_unix_server_open(self):
@@ -924,7 +1067,7 @@ class _TestUNIXForwarding(_CheckForwarding):
             async with conn.start_unix_server(_unix_listener_non_async, path):
                 await self._check_local_unix_connection('echo')
 
-        os.remove('echo')
+        try_remove('echo')
 
     @asynctest
     async def test_unix_server_failure(self):
@@ -942,7 +1085,65 @@ class _TestUNIXForwarding(_CheckForwarding):
             async with conn.forward_local_path('local', '/echo'):
                 await self._check_local_unix_connection('local')
 
-        os.remove('local')
+        try_remove('local')
+
+    @asynctest
+    async def test_forward_local_port_to_path_accept_handler(self):
+        """Test forwarding of port to UNIX path with accept handler"""
+
+        def accept_handler(_orig_host: str, _orig_port: int) -> bool:
+            return True
+
+        async with self.connect() as conn:
+            async with conn.forward_local_port_to_path(
+                    '', 0, '/echo', accept_handler) as listener:
+                await self._check_local_connection(listener.get_port(),
+                                                   delay=0.1)
+
+    @asynctest
+    async def test_forward_local_port_to_path_accept_handler_denial(self):
+        """Test forwarding of port to UNIX path with accept handler denial"""
+
+        async def accept_handler(_orig_host: str, _orig_port: int) -> bool:
+            return False
+
+        async with self.connect() as conn:
+            async with conn.forward_local_port_to_path(
+                    '', 0, '/echo', accept_handler) as listener:
+                listen_port = listener.get_port()
+
+                reader, writer = await asyncio.open_connection('127.0.0.1',
+                                                               listen_port)
+
+                self.assertEqual((await reader.read()), b'')
+
+                writer.close()
+                await maybe_wait_closed(writer)
+
+    @asynctest
+    async def test_forward_local_port_to_path(self):
+        """Test forwarding of a local port to a remote UNIX domain socket"""
+
+        async with self.connect() as conn:
+            async with conn.forward_local_port_to_path('', 0,
+                                                       '/echo') as listener:
+                await self._check_local_connection(listener.get_port(),
+                                                   delay=0.1)
+
+    @asynctest
+    async def test_forward_specific_local_port_to_path(self):
+        """Test forwarding of a specific local port to a UNIX domain socket"""
+
+        sock = socket.socket()
+        sock.bind(('', 0))
+        listen_port = sock.getsockname()[1]
+        sock.close()
+
+        async with self.connect() as conn:
+            async with conn.forward_local_port_to_path(
+                    '', listen_port, '/echo') as listener:
+                await self._check_local_connection(listener.get_port(),
+                                                   delay=0.1)
 
     @asynctest
     async def test_forward_remote_path(self):
@@ -962,8 +1163,28 @@ class _TestUNIXForwarding(_CheckForwarding):
         server.close()
         await server.wait_closed()
 
-        os.remove('echo')
-        os.remove('local')
+        try_remove('echo')
+        try_remove('local')
+
+    @asynctest
+    async def test_forward_remote_path_to_port(self):
+        """Test forwarding of a remote UNIX domain path to a local TCP port"""
+
+        server = await asyncio.start_server(echo, None, 0,
+                                            family=socket.AF_INET)
+        server_port = server.sockets[0].getsockname()[1]
+
+        path = os.path.abspath('echo')
+
+        async with self.connect() as conn:
+            async with conn.forward_remote_path_to_port(
+                    path, '127.0.0.1', server_port):
+                await self._check_local_unix_connection('echo')
+
+        server.close()
+        await server.wait_closed()
+
+        try_remove('echo')
 
     @asynctest
     async def test_forward_remote_path_failure(self):
@@ -977,7 +1198,7 @@ class _TestUNIXForwarding(_CheckForwarding):
             with self.assertRaises(asyncssh.ChannelListenError):
                 await conn.forward_remote_path(path, 'local')
 
-        os.remove('echo')
+        try_remove('echo')
 
     @asynctest
     async def test_forward_remote_path_not_permitted(self):
